# -*- coding: utf-8 -*-
"""
Functions to delineate catchments.
"""
import numpy as np
import pandas as pd
from gistools.vector import kd_nearest, convert_lines_to_points
from gistools.util import load_geo_data
from shapely.geometry import Polygon, Point, box
from shapely.ops import unary_union

#####################################################
#### MFE REC streams network


def find_upstream(nzreach, rec_streams, segment_id_col='nzsegment', from_node_col='FROM_NODE', to_node_col='TO_NODE'):
    """
    Function to estimate all of the reaches (and nodes) upstream of specific reaches.

    Parameters
    ----------
    nzreach : list, ndarray, Series of int
        The NZ reach IDs
    rec_streams_shp : str path or GeoDataFrame
        str path to the REC streams shapefile or the equivelant GeoDataFrame.
    segment_id_col : str
        The column name of the line segment id.
    from_node_col : str
        The from node column
    to_node_col : str
        The to node column

    Returns
    -------
    DataFrame
    """
    if not isinstance(nzreach, (list, np.ndarray, pd.Series)):
        raise TypeError('nzreach must be a list, ndarray or Series.')

   ### Load data
    rec = load_geo_data(rec_streams).drop('geometry', axis=1).copy()

    ### Run through all nzreaches
    reaches_lst = []
    for i in nzreach:
        reach1 = rec[rec[segment_id_col] == i].copy()
        up1 = rec[rec[to_node_col].isin(reach1[from_node_col])]
        while not up1.empty:
            reach1 = pd.concat([reach1, up1])
            up1 = rec[rec[to_node_col].isin(up1[from_node_col])]
        reach1.loc[:, 'start'] = i
        reaches_lst.append(reach1)

    reaches = pd.concat(reaches_lst)
    reaches.set_index('start', inplace=True)

    return reaches


###############################################
### Catch delineation using the REC


def extract_catch(reaches, rec_catch, segment_id_col='nzsegment'):
    """
    Function to extract the catchment polygons from the rec catchments layer. Appends to reaches layer.

    Parameters
    ----------
    reaches : DataFrame
        The output DataFrame from the find_upstream function.
    rec_catch_shp : str path, dict, or GeoDataFrame
        str path to the REC catchment shapefile or a GeoDataFrame.
    segment_id_col : str
        The column name of the line segment id.

    Returns
    -------
    GeoDataFrame
    """
    sites = reaches[segment_id_col].unique().astype('int32')
    catch0 = load_geo_data(rec_catch)

    catch1 = catch0[catch0[segment_id_col].isin(sites)].copy()
    catch2 = catch1.dissolve(segment_id_col).reset_index()[[segment_id_col, 'geometry']]

    ### Combine with original sites
    catch3 = catch2.merge(reaches.reset_index(), on=segment_id_col)
    catch3.crs = catch0.crs

    return catch3


def agg_catch(rec_catch):
    """
    Simple function to aggregate rec catchments.

    Parameters
    ----------
    rec_catch : GeoDataFrame
        The output from extract_catch

    Returns
    -------
    GeoDataFrame
    """
    rec_shed = rec_catch[['start', 'geometry']].dissolve('start')
    rec_shed.index = rec_shed.index.astype('int32')
    rec_shed['area'] = rec_shed.area
    rec_shed.crs = rec_catch.crs
    return rec_shed.reset_index()


def catch_delineate(sites, rec_streams, rec_catch, segment_id_col='nzsegment', from_node_col='FROM_NODE', to_node_col='TO_NODE', ignore_order=1, stream_order_col='StreamOrde', max_distance=np.inf, site_delineate='all', returns='catch'):
    """
    Catchment delineation using the REC streams and catchments.

    Parameters
    ----------
    sites : str path or GeoDataFrame
        Points shapfile of the sites along the streams or the equivelant GeoDataFrame.
    rec_streams : str path or GeoDataFrame
        str path to the REC streams shapefile, the equivelant GeoDataFrame, or a dict of parameters to read in an mssql table using the rd_sql function.
    rec_catch : str path or GeoDataFrame
        str path to the REC catchment shapefile, the equivelant GeoDataFrame, or a dict of parameters to read in an mssql table using the rd_sql function.
    segment_id_col : str
        The column name of the line segment id.
    from_node_col : str
        The from node column
    to_node_col : str
        The to node column
    ignore_order : int
        Ignore the stream orders in the search up to this int.
    stream_order_col : str
        The stream order column.
    max_distance : non-negative float, optional
        Return only neighbors within this distance. This is used to prune tree searches, so if you are doing a series of nearest-neighbor queries, it may help to supply the distance to the nearest neighbor of the most recent point. It's best to define a reasonable distance for the search.
    site_delineate : 'all' or 'between'
        Whether the catchments should be dileated all the way to the top or only in between the sites.
    returns : 'catch' or 'all'
        Return only the catchment polygons or the catchments, reaches, and sites

    Returns
    -------
    GeoDataFrame
        Polygons
    """

    ### Parameters


    ### Modifications {segment_id_col: {NZTNODE/NZFNODE: node # to change}}
    # mods = {13053151: {segment_id_col: 13055874}, 13048353: {'NZTNODE': 13048851}, 13048498: {'NZTNODE': 13048851}, 13048490: {'ORDER': 3}}

    ### Load data
    rec_catch = load_geo_data(rec_catch)
    rec_streams = load_geo_data(rec_streams)
    pts = load_geo_data(sites)
    pts['geometry'] = pts.geometry.simplify(1)

    ### make mods
    # for i in mods:
    #     rec_streams.loc[rec_streams['segment_id_col'] == i, list(mods[i].keys())] = list(mods[i].values())

    ### Find closest REC segment to points
    if max_distance == np.inf:
        buffer_dis = 100000
    else:
        buffer_dis = max_distance

    pts_extent = box(*pts.unary_union.buffer(buffer_dis).bounds)

    s_order = list(range(1, ignore_order + 1))
    rec_streams2 = rec_streams[~rec_streams[stream_order_col].isin(s_order)]

    rec_pts2 = convert_lines_to_points(rec_streams2, segment_id_col, pts_extent)

    # rec_pts1 = rec_streams2[rec_streams2.intersects(pts_extent)].set_index(segment_id_col).copy()
    # coords = rec_pts1.geometry.apply(lambda x: list(x.coords)).explode()
    # geo1 = coords.apply(lambda x: Point(x))
    #
    # rec_pts2 = gpd.GeoDataFrame(coords, geometry=geo1, crs=rec_pts1.crs).reset_index()

    pts_seg = kd_nearest(pts, rec_pts2, segment_id_col, max_distance=max_distance)
    nzreach = pts_seg[segment_id_col].copy().unique()

    ### Find all upstream reaches
    reaches = find_upstream(nzreach, rec_streams=rec_streams, segment_id_col=segment_id_col, from_node_col=from_node_col, to_node_col=to_node_col)

    ### Clip reaches to in-between sites if required
    if site_delineate == 'between':
        reaches1 = reaches.reset_index().copy()
        reaches2 = reaches1.loc[reaches1[segment_id_col].isin(reaches1.start.unique()), ['start', segment_id_col]]
        reaches2 = reaches2[reaches2.start != reaches2[segment_id_col]]

        grp1 = reaches2.groupby('start')

        for index, r in grp1:
#            print(index, r)
            r2 = reaches1[reaches1.start.isin(r[segment_id_col])][segment_id_col].unique()
            reaches1 = reaches1[~((reaches1.start == index) & (reaches1[segment_id_col].isin(r2)))]

        reaches = reaches1.set_index('start').copy()

    ### Extract associated catchments
    rec_catch2 = extract_catch(reaches, rec_catch=rec_catch, segment_id_col=segment_id_col)

    ### Aggregate individual catchments
    rec_shed = agg_catch(rec_catch2)
    rec_shed.columns = [segment_id_col, 'geometry', 'area']
    rec_shed1 = rec_shed.merge(pts_seg.drop('geometry', axis=1), on=segment_id_col)

    ### Return
    if returns == 'catch':
        return rec_shed1
    else:
        return rec_shed1, reaches, pts_seg



# -*- coding: utf-8 -*-
"""
Created on Mon Dec 23 14:36:11 2019

@author: MichaelEK
"""
import numpy as np
import pandas as pd
import geopandas as gpd
from scipy.spatial import cKDTree
from gistools import vector
import copy
import time
try:
    import overpass
    import osm2geojson
except:
    print('Install overpass and osm2geojson for osm module functions')

#############################################
### Parameters

op_endpoint = 'https://overpass-api.de/api/interpreter'


#############################################
### Functions


def get_nearest_waterways(gdf_from, id_col, max_distance=500, waterway_type='all'):
    """
    Function to get the nearest waterways and associated nodes at some max distance from the OSM overpass server. Can specify the waterway type as either 'natural' or 'all'. The higher the max distance the slower the server query.

    Parameters
    ----------
    gdf_from : GeoDataFrame
        Locations of the sites where you want to find the nearest waterway nodes.
    id_col : str
        The column name of the site ID for gdf_from.
    max_distance : int
        The maximum search distance.
    waterway_type : str
        Either 'nautral' or 'all'

    Returns
    -------
    GeoDataFrame
        of the sites and associated waterys and nodes.
    """
    if waterway_type == 'all':
        q_base = """(way['waterway'~'(river|stream|drain|canal|tidal_channel|ditch)'](around:{dis}, {lat}, {lon});
    node(around:{dis}, {lat}, {lon})(w);)"""
    elif waterway_type == 'natural':
        q_base = "(way['waterway'~'(river|stream|tidal_channel)'](around:{dis}, {lat}, {lon}); node(around:{dis}, {lat}, {lon})(w);)"
    else:
        raise ValueError('waterway_type must be either natural or all.')

    from1 = gdf_from[[id_col, 'geometry']].copy()

    pts1 = from1.to_crs(4326)

    from1['x'] = from1.geometry.x
    from1['y'] = from1.geometry.y
    from1['lon'] = pts1.geometry.x
    from1['lat'] = pts1.geometry.y

    api = overpass.API(endpoint=op_endpoint)

    print('Overpass endpoint is: ' + op_endpoint)

    no_node_ids = []

    res_list = []
    for index, p in from1.iterrows():
#        print(p[id_col])
        q1 = q_base.format(dis=max_distance, lat=p.lat, lon=p.lon)

        response = api.get(q1, responseformat='json')
        time.sleep(1)

        nodes1 = [n for n in response['elements'] if n['type'] == 'node']
        pd_nodes = pd.DataFrame.from_records(nodes1)
        if pd_nodes.empty:
            print('No node was found during the query for ' + id_col + ': ' + str(p[id_col]))
            no_node_ids.extend([p[id_col]])
            continue
        gpd_nodes1 = gpd.GeoDataFrame(pd_nodes, geometry=gpd.points_from_xy(pd_nodes['lon'], pd_nodes['lat']), crs=4326)
        gpd_nodes2 = gpd_nodes1.to_crs(from1.crs)

        nodes2 = list(zip(gpd_nodes2.geometry.x, gpd_nodes2.geometry.y))

        btree = cKDTree(nodes2)
        dist, idx = btree.query([p.x, p.y], k=1, distance_upper_bound=max_distance)

        best1 = gpd_nodes2.iloc[[idx]].copy()
        best1['distance'] = round(dist, 2)
        best1[id_col] = p[id_col]

        ways1 = [n for n in response['elements'] if n['type'] == 'way']
        [s['tags'].update({'name': 'No name'}) for s in ways1 if not 'name' in s['tags']]
        if len(ways1) == 1:
            this_way = {'waterway_id': ways1[0]['id'], 'waterway_name': ways1[0]['tags']['name']}
        else:
            this_way = [{'waterway_id': w['id'], 'waterway_name': w['tags']['name']} for w in ways1 if best1['id'].iloc[0] in w['nodes'][1:]][0]
        best1['waterway_id'] = this_way['waterway_id']
        best1['waterway_name'] = this_way['waterway_name']

        res_list.append(best1)

    res1 = pd.concat(res_list, sort=False).reset_index(drop=True)
    return res1, no_node_ids


def get_waterways(osm_nodes_from, waterway_type='all'):
    """
    Function to get all the waterways connected by the node IDs derived from the get_nearest_waterways function from the OSM overpass server. Can specify the waterway type as either 'natural' or 'all'.

    Parameters
    ----------
    osm_nodes_from : GeoDataFrame
        Output GeoDataFrame from the get_nearest_waterways function with waterway nodes.
    waterway_type : str
        Either 'nautral' or 'all'.

    Returns
    -------
    Two Dicts
        of waterways and nodes.
    """
    q_node_base = "node({node});"

    if waterway_type == 'all':
        q_other_base = """complete {(way['waterway'](<);
                            (>;);
                            way['natural'='water'](<);
                            (>;););}"""
    elif waterway_type == 'natural':
        q_other_base = """complete {(way['waterway'~'(river|stream|tidal_channel)'](<);
                            (>;);
                            way['natural'='water'](<);
                            (>;););}"""
    else:
        raise ValueError('waterway_type must be either natural or all.')

    api = overpass.API(endpoint=op_endpoint)

    waterways = []
    nodes = {}

    for index, p in osm_nodes_from.iterrows():
        w_check = True
        for w in waterways:
            if p.waterway_id in w[1].keys():
                w[0].update({p.id: p.waterway_id})
                w_check = False
                break
        if w_check:
            print(p)
            q_node = q_node_base.format(node=p.id)
            q1 = q_node + q_other_base

            response = api.get(q1, responseformat='json')
            time.sleep(1)

            ww1 = {ww['id']: ww for ww in response['elements'] if (ww['type'] == 'way')}
            lst1 = [{p.id: p.waterway_id}, ww1]
            waterways.append(lst1)
            n1 = {n['id']: n for n in response['elements'] if n['type'] == 'node'}
            nodes.update(n1)

    return waterways, nodes


def waterway_delineation(waterways, site_delineate=False, only_waterways=True):
    """
    Function to delineate the waterways above each of the sites/nodes originally derived by the get_nearest_waterways function. Optional to delineate all the way up the waterway network or between the sites.

    Parameters
    ----------
    osm_nodes_from : GeoDataFrame
        Output GeoDataFrame from the get_nearest_waterways function with waterway nodes.
    waterways : dict
        First output dict from the get_waterways function.
    site_delineate : bool
        Whether the waterway network should be dileated all the way to the top (False) or only in between the sites (True).

    Returns
    -------
    Dict
        of each site/node each continaing the associated waterways and nodes.
    """
    site_delin = {}
    for w in waterways:
        ways = w[1]
        for node, node_way in w[0].items():
            new_wws = copy.deepcopy(ways)

            site_ww = {node_way: ways[node_way].copy()}
            site_ww_nodes1 = site_ww[node_way]['nodes']
            site_node_index = site_ww_nodes1.index(node)+1
            if (site_node_index + 1) == len(site_ww_nodes1):
                site_ww_nodes = site_ww_nodes1
            else:
                site_ww_nodes = site_ww_nodes1[:site_node_index]

            site_ww[node_way]['nodes'] = site_ww_nodes

            ww_last = {}
            for i, ww in new_wws.items():
                if 'waterway' in ww['tags']:
                    ww_last.update({i: [ww['nodes'][-1]]})
                else:
                    ww_last.update({i: ww['nodes']})

            if len(site_ww_nodes1) == site_node_index:
                big_set = set(site_ww_nodes[:-1])
            else:
                big_set = set(site_ww_nodes)

            set_len = len(big_set)

            bigger_set = big_set.copy()

            while set_len > 0:
                index1 = [i for i, ww in ww_last.items() if np.in1d(ww, list(big_set)).any()]
                new_ww = {i: new_wws[i] for i in index1}
                site_ww.update({i: new_ww[i] for i in index1})
                bigger_set.update(big_set)
                temp_set = set()
                [temp_set.update(set(new_ww[i]['nodes'][:-1])) for i in index1]

                big_set = temp_set.difference(bigger_set)

                set_len = len(big_set)

            if only_waterways:
                site_ww1 = {i: s for i, s in site_ww.items() if 'waterway' in s['tags']}
                site_ww2 = {i: s for i, s in site_ww1.items() if s['tags']['waterway'] in ['river', 'stream', 'drain', 'canal', 'ditch']}
            else:
                site_ww2 = site_ww

            site_delin.update({node: site_ww2})

        if site_delineate:
            for node1, node_way1 in w[0].items():
#                print('origin ' + str(node1))
                sw1 = site_delin[node1]
                for node2, node_way2 in w[0].items():
                    if node2 != node1:
#                        print('removal ' + str(node2))
                        if node_way2 in sw1.keys():
                            sw2 = site_delin[node2]
                            way1 = sw1[node_way2]['nodes']
                            if (node2 in way1) and (node_way2 in sw2):
                                other_way_ids = list(sw2.keys())
                                other_way2 = sw2[node_way2]['nodes']
                                if (len(other_way2)+1) < len(way1):
                                    new_way1 = way1[(way1.index(node2)):]
                                    site_delin[node1][node_way2]['nodes'] = new_way1
                                    other_way_ids.remove(node_way2)
                                [site_delin[node1].pop(w, None) for w in other_way_ids]

    return site_delin


def to_osm(site_delin, nodes):
    """
    Function to convert the output of the waterway_delineation function to a dict of dict in the OSM dict structure (for futher processing).

    Parameters
    ----------
    site_delin : dict
        Output of the waterway_delineation function.
    nodes : dict
        Second output of the get_waterways function.

    Returns
    -------
    Dict
        dict of dict in the OSM dict structure.
    """
    time1 = pd.Timestamp.now().isoformat()

    osm_delin = {}

    for id1, ww in site_delin.items():
        big_set = set()
        [big_set.update(set(v['nodes'])) for v in ww.values()]
        lotsanodes = [n for i, n in nodes.items() if i in big_set]
        lotsaways = [l for i, l in ww.items()]

        lots = []
        lots.extend(lotsanodes)
        lots.extend(lotsaways)

        dict1 = {'timestamp': time1, 'elements': lots}
        osm_delin.update({id1: dict1})

    return osm_delin


def to_gdf(osm_delin):
    """
    Function to convert the output of the to_osm function to a GeoDataFrame of the waterways associated with the sites/nodes.

    Parameters
    ----------
    osm_delin : dict
        Output of the to_osm function.

    Returns
    -------
    GeoDataFrame
    """
    shape1 = []

    for id1, osm1 in osm_delin.items():
        s1 = osm2geojson.json2shapes(osm1)
        [s['properties']['tags'].update({'name': 'No name'}) for s in s1 if not 'name' in s['properties']['tags']]
        l1 = [[id1, s['properties']['id'], s['properties']['tags']['name'], s['properties']['tags']['waterway'], s['shape']] for s in s1]
        shape1.extend(l1)

    df1 = pd.DataFrame(shape1, columns=['start_node', 'way_id', 'name', 'waterway', 'geometry'])
#    gdf1 = gpd.GeoDataFrame(df1, crs=4326, geometry='geometry').dissolve(['start_node', 'way_id']).reset_index()
    gdf1 = gpd.GeoDataFrame(df1, crs='epsg:4326', geometry='geometry')
    gdf2 = gdf1[gdf1.geom_type == 'LineString'].copy()

    return gdf2


def to_nx():
    """
    To be completed...convert to networkx. Look at osmnx...
    """


def pts_to_waterway_delineation(gdf_from, id_col, max_distance=500, waterway_type='all', site_delineate=False):
    """
    Function to fully perform the OSM waterway delineation process.

    Parameters
    ----------
    Parameters
    ----------
    gdf_from : GeoDataFrame
        Locations of the sites where you want to find the nearest waterway nodes.
    id_col : str
        The column name of the site ID for gdf_from.
    max_distance : int
        The maximum search distance.
    waterway_type : str
        Either 'nautral' or 'all'
    site_delineate : bool
        Whether the waterway network should be dileated all the way to the top (False) or only in between the sites (True).

    Returns
    -------
    Two GeoDataFrames
        The first is the GeoDataFrame of the nearest OSM waterway and node associated with the sites and the other is the one that contains the upstream delineated waterways.
    """
    pts1 = get_nearest_waterways(gdf_from, id_col, max_distance, waterway_type)
    waterways, nodes = get_waterways(pts1, waterway_type)
    site_delin = waterway_delineation(waterways, site_delineate)
    osm_delin = to_osm(site_delin, nodes)
    gdf1 = to_gdf(osm_delin).to_crs(pts1.crs)

    return pts1, gdf1


def get_waterways_within_boundary(poly, buffer=0, waterway_type='all'):
    """

    """
    if waterway_type == 'all':
        q_base = """way['waterway']({min_lat}, {min_lon}, {max_lat}, {max_lon});(._;>;);"""
    elif waterway_type == 'natural':
        q_base = """(way['waterway'~'(river|stream|tidal_channel)']({min_lat}, {min_lon}, {max_lat}, {max_lon});(._;>;);"""
    else:
        raise ValueError('waterway_type must be either natural or all.')

    ## Prepare the polygon
    poly1 = poly.to_crs('epsg:4326').copy()
    poly1['geometry'] = poly1.buffer(buffer)
    poly2 = poly1.unary_union
    min_lon, min_lat, max_lon, max_lat = poly2.bounds

    ## Query OSM
    q1 = q_base.format(min_lon=min_lon, min_lat=min_lat, max_lon=max_lon, max_lat=max_lat)

    api = overpass.API(endpoint=op_endpoint)

    print('Overpass endpoint is: ' + op_endpoint)

    response = api.get(q1, responseformat='json')

    ## convert to gpd
    s1 = osm2geojson.json2shapes(response)
    [s['properties']['tags'].update({'name': 'No name'}) for s in s1 if not 'name' in s['properties']['tags']]
    l1 = [[s['properties']['id'], s['properties']['tags']['name'], s['properties']['tags']['waterway'], s['shape']] for s in s1]
    df1 = pd.DataFrame(l1, columns=['way_id', 'name', 'waterway', 'geometry'])
    gdf1 = gpd.GeoDataFrame(df1, crs='epsg:4326', geometry='geometry')
    gdf2 = gdf1[gdf1.geom_type == 'LineString'].copy()

    ## Select rivers within polygon
    gdf3 = gdf2[gdf2.intersects(poly2)].to_crs(poly.crs).copy()

    ## Return
    return gdf3













